# coding=utf8
import sys
import random
import subprocess
from os import popen

from PIL import Image, ImageDraw

from printer.font_helper import initiate_true_type

DEFAULT_SIZE = 50, 30  # width, height
_SIZE_CMD = "stty size"

if sys.version_info.major == 2:
    reload(sys)
    sys.setdefaultencoding('utf8')
    MESS_FILTERS = ''.join([unichr(i) for i in range(32, 256)])
else:
    MESS_FILTERS = ''.join([chr(i) for i in range(32, 256)])
    unicode = lambda s: s

try:
    _code, _output = subprocess.getstatusoutput(_SIZE_CMD)
    if _code == 0:
        DEFAULT_SIZE = [int(s) for s in _output.split(' ')][::-1]
except AttributeError:  # python2
    try:
        DEFAULT_SIZE = [int(s) for s in popen(_SIZE_CMD).read().split(' ')][::-1]
    except:
        pass


class ImageMap(object):
    def __init__(self):
        self._image = {}

    def __setitem__(self, key, value):
        self._image[key] = value

    def __getitem__(self, key):
        return self._image[key]


def make_terminal_img(img, filter_type=None, width=None,
                      height=None, dye=None, reverse=False,
                      keep_ratio=False, gray=True, strip_white=False):
    """
    在终端输出字符图片
    :param img: 图像
    :param filter_type: 填充字符
    :param width: 输出宽度
    :param height: 输出高度
    :param dye: 上色
    :param reverse: 是否反色
    :param keep_ratio: 是否保持比例
    :param gray: 如果处理图片，是否转换为灰度图
    :param strip_white: 是否删除(文字打印)模式下的空白行
    :return: 图像字符
    """
    if not img:
        # 如果文字画布生成失败，img为空
        return ''
    if not keep_ratio:
        img = img.resize((width or DEFAULT_SIZE[0], height or DEFAULT_SIZE[1]))
    else:
        size = img.size
        if width is None or height is None:
            ratio = min(float(DEFAULT_SIZE[0]) / size[0], float(DEFAULT_SIZE[1]) / size[1])
            img = img.resize((int(size[0] * ratio), int(size[1] * ratio)))
        else:
            ratio = min(float(width) / size[0], float(height) / size[1])
            img = img.resize((int(size[0] * ratio), int(size[1] * ratio)))
    width, height = img.size
    pix = img.load()

    if strip_white:
        image_map = ImageMap()
        white_lines = 0
        after_h = 0
        is_start_line = True
        for h in range(height):
            if all([pix[w, h] == 255 for w in range(width)]):
                if is_start_line:
                    # skip first white line
                    is_start_line = False
                elif h == height - 2:
                    # skip last white line
                    pass
                else:
                    white_lines += 1
                    continue
            for w in range(width):
                image_map[w, after_h] = pix[w, h]
            after_h += 1
        height = after_h
        pix = image_map

    if gray:
        def render_pix(x, y):
            if filter_type:
                if reverse:
                    return MESS_FILTERS[(255 - pix[x, y]) * filter_type // 255]
                return MESS_FILTERS[pix[x, y] * filter_type // 255]
            else:
                if reverse:
                    return MESS_FILTERS[(255 - pix[x, y]) * 4 // 255]
                return MESS_FILTERS[pix[x, y] * 4 // 255]
    else:
        def render_pix(x, y):
            # 如果这里也用和灰度图一样的处理方法的话，会显得很乱，终端中的颜色也难以显示出来
            return '\033[0;38;2;%s;%s;%sm' % pix[x, y][:3] + MESS_FILTERS[filter_type]

    if type(dye) is int:
        # 特定颜色绘制
        result = '\033[01;{}m'.format(dye) + \
                 '\n'.join([''.join([render_pix(w, h) for w in range(width)])
                            for h in range(height)]) + '\033[00m'
    elif type(dye) is str:
        # 随机颜色绘制
        result = '\n'.join([''.join(["\033[01;{}m{}".format(random.randrange(30, 40),
                                                            render_pix(w, h))
                                     for w in range(width)])
                            for h in range(height)]) + '\033[00m'

    else:
        # 黑白
        result = '\n'.join([''.join([render_pix(w, h)
                                     for w in range(width)])
                            for h in range(height)])
    img.close()
    return result


def get_img(file_path, gray=False):
    """
    获取输入图片
    :param file_path: 图片文件位置
    :param gray: 是否转换为灰度图
    :return: img
    """
    try:
        img = Image.open(file_path)
        if gray:
            img = img.convert('L')
        return img
    except Exception:
        print("不支持的图片格式")
        return None


def text_drawer(text, fonts=None):
    """
    将文字书写在白色画布上
    :param text: 要书写的文字
    :param fonts: 字体选择，索引或路径
    :return: img
    """
    im = Image.new("1", (1, 1), 'white')  # 初始画布大小没有关系
    draw = ImageDraw.Draw(im)
    font = initiate_true_type(fonts, 20)
    if not font:
        return None
    text_size = draw.textsize(unicode(text), font=font)
    im = im.resize(text_size)

    draw = ImageDraw.Draw(im)
    draw.text((0, 0), unicode(text), font=font)
    return im




# coding=utf8

import os
import sys
import time
import shutil
from os import path

from PIL import ImageFont

from printer.http import HTTPCons, SockFeed, unit_change

if sys.version_info.major == 2:
    reload(sys)
    sys.setdefaultencoding("utf8")

# __all__ = ['font_downloader', 'font_handle']

FONT_LIST = ['shuyan.ttf',
             'letter.ttf',
             'Haibaoyuanyuan.ttf',
             'fengyun.ttf',
             'huakangbold.otf']

_font_prefix = "https://raw.githubusercontent.com/hellflame/terminal_printer/" \
               "808004a7cd41b4383bfe6aa310c491c69d9b2556/fonts/"

FONT_URL = {
    f: _font_prefix + f for f in FONT_LIST
}

FONT_DIR = path.join(path.expanduser('~'), ".terminal_fonts")


def choose_font(choice):
    if choice.isdigit():
        choice = int(choice)
        font = path.join(FONT_DIR, FONT_LIST[choice if len(FONT_LIST) - 1 >= choice >= 0 else 0])
    else:
        font = choice
    if not path.exists(font) and path.isfile(font):
        return font, False
    return font, True


def initiate_true_type(choice, size=20):
    font_path, exist = choose_font(choice)
    if not exist:
        print("字体文件不存在({})，请使用其他字体".format(font_path))
        return None
    try:
        return ImageFont.truetype(font_path, size)
    except IOError:
        print("字体文件损坏({})，请使用其他字体".format(font_path))
        return None


def font_downloader(font_link, font_dir):
    """
    字体下载
    :param font_link: 字体名称
    :param font_dir: 字体保存路径
    :return:
    """
    font_name = os.path.basename(font_link)
    save_path = os.path.join(font_dir, font_name)
    downloader = HTTPCons()
    downloader.request(font_link)
    feed = SockFeed(downloader)
    start = time.time()
    feed.http_response(save_path, chunk=4096)
    end = time.time()
    if int(feed.status['code']) == 200:
        size = os.stat(save_path).st_size
        print("\033[01;31m{}\033[00m downloaded @speed \033[01;32m{}/s\033[00m"
              .format(font_name,
                      unit_change(size / (end - start))))
    else:
        print("\033[01;31m{}\033[00m 下载失败".format(font_name))
    return True


def font_init(show_prompt=True):
    """
    字体下载管理，如果没有缺失字体依然执行，将提示重新下载所有字体
    :param show_prompt: 显示提示信息
    :return:
    """
    target = [FONT_URL[f] for f in FONT_URL if not os.path.exists(os.path.join(FONT_DIR, f))]
    if not target:
        # 如果字体完整依然执行初始化，则提示删除原有字体目录
        if show_prompt:
            # 如果不显示提示信息，则直接删除
            prompt = "当前字体数据完整，是否删除后继续初始化? y/n "
            try:
                if sys.version_info.major == 2:
                    if not raw_input(prompt).lower().startswith('y'):
                        return False
                else:
                    if not input(prompt).lower().startswith('y'):
                        return False
            except KeyboardInterrupt:
                exit(0)
        shutil.rmtree(FONT_DIR)
        target = FONT_URL.values()

    if not os.path.exists(FONT_DIR):
        # 创建字体目录
        os.makedirs(FONT_DIR)

    print("Start Downloading {} fonts".format(len(target)))
    for font in target:
        font_downloader(font, FONT_DIR)

    print("下载完成")




# coding=utf8

from setuptools import setup
from printer.version import __version__, __author__, __url__

setup(
    name='TerminalPrinter',
    version=__version__,
    keywords=('字符画', '终端打印'),
    description="终端图片、文字生成器",
    license='MIT',
    author=__author__,
    author_email='hellflamedly@gmail.com',
    url=__url__,
    packages=["printer"],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.10',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        "Operating System :: MacOS",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        'Topic :: Text Processing :: General',
        'Topic :: Terminals',
        'Topic :: Games/Entertainment'
    ],
    install_requires=[
        'Pillow'
    ],
    entry_points={
        'console_scripts': [
            'terminalprint=printer.run:run'
        ]
    }
)





# coding=utf8

import os
import argparse

from printer.version import __url__
from printer.painter import MESS_FILTERS, make_terminal_img, text_drawer, get_img
from printer.font_helper import font_init, choose_font
from printer.utils import print_version, print_debug

__all__ = ['parser']


def command(args, parse):
    if args.debug:
        print_debug(args)
    if args.init:
        font_init()
    elif args.version:
        print_version()
    elif args.picture:
        pic_path = args.picture
        if os.path.exists(pic_path) and os.path.isfile(pic_path):
            print(make_terminal_img(
                get_img(pic_path, gray=args.gray),
                filter_type=args.filter, width=args.width, height=args.height,
                dye=args.color, reverse=args.reverse, keep_ratio=args.keep_ratio,
                gray=args.gray))
        else:
            print("请输入有效图片路径")
    elif args.text:
        print(make_terminal_img(
            text_drawer(args.text, args.font),
            filter_type=args.filter, width=args.width, height=args.height,
            dye=args.color, reverse=not args.reverse, keep_ratio=False, gray=True,
            strip_white=True))
    else:
        parse.print_help()


def parser():
    parse = argparse.ArgumentParser(description="Terminal Printer",
                                    formatter_class=argparse.RawTextHelpFormatter,
                                    epilog="首次进行文字处理\r\n"
                                           "需要执行 terminalprint -i 初始化或指定字体\r\n"
                                           "更多帮助信息请参考: " + __url__)

    def usable_color(s):
        _COLOR_MAP = {'black': 30, 'red': 31, 'green': 32, 'yellow': 33, 'blue': 34,
                      'magenta': 35, 'cyan': 36, 'white': 37}
        for k in tuple(_COLOR_MAP.keys()):
            _COLOR_MAP['bg-' + k] = _COLOR_MAP[k] + 10
        if s.isdigit():
            if 30 <= int(s) <= 50:
                return int(s)
            raise argparse.ArgumentTypeError("颜色值若为数字，应在30～50之间")
        else:
            return _COLOR_MAP.get(s, s)

    def usable_filter(s):
        if s.isdigit() and 1 <= int(s) <= len(MESS_FILTERS) - 1:
            return int(s)
        raise argparse.ArgumentTypeError("填充方式索引值应在1～{}之间".format(len(MESS_FILTERS) - 1))

    def usable_font(s):
        f, exist = choose_font(s)
        if not exist:
            raise argparse.ArgumentTypeError("字体路径不存在，请检查路径或使用数字")
        return f

    basic = parse.add_argument_group("basics")
    basic.add_argument("-i", "--init", action="store_true", help="初始化程序，下载字体")
    basic.add_argument("-v", '--version', action="store_true", help="输出版本信息")
    basic.add_argument("--debug", action="store_true", help="输出调试信息")

    picture = parse.add_argument_group("pictures")
    picture.add_argument("picture", nargs="?", help="可选的图片")
    picture.add_argument("-kr", '--keep-ratio', action="store_true", help="保持图片比例")

    text = parse.add_argument_group("text")
    text.add_argument("-t", '--text', default="HellFlame", help="设置将要处理的文本内容，默认为 HellFlame")
    text.add_argument("-c", '--color', type=usable_color, metavar="i", help="设置颜色")
    text.add_argument("-g", '--gray', action="store_true", help="图像转换为灰度图(若指定图)")
    text.add_argument("-F", '--font', metavar="path", type=usable_font, default='0', help="设置书写字体")
    text.add_argument("-r", '--reverse', action="store_true", help="反色(对彩色输出无效)")

    common = parse.add_argument_group("common")
    common.add_argument('-W', "--width", metavar="w", type=int, help="设置输出宽度，需要与高度一起设置")
    common.add_argument('-H', "--height", metavar="h", type=int, help="设置输出高度，需要与宽度一起设置")
    common.add_argument("-f", '--filter', type=usable_filter, metavar="i", default=73, help="设置打印填充方式")

    # 可选的位置参数
    return parse.parse_args(), parse


def run():
    command(*parser())


if __name__ == '__main__':
    run()




import os
import platform

from printer.version import __version__
from printer.painter import DEFAULT_SIZE
from printer.font_helper import choose_font


def print_version():
    print("TerminalPrinter v{}".format(__version__))


def print_debug(args):
    print_version()
    print("Shell: {}".format(os.getenv("SHELL")))
    print("Term: {}".format(os.getenv("TERM")))
    print("Platform: {}".format("/".join(platform.uname())))
    print("Given Size: {}/{} Default Size: {}/{}".format(args.width, args.height, *DEFAULT_SIZE))
    print("Arguments: {}".format(" ".join(["{}:{}".format(k, v) for k, v in args.__dict__.items()])))
    if args.text:
        font, exist = choose_font(args.font)
        print("Font: {} (exist? {})".format(font, exist))



# coding=utf8

import socket
import math
import sys
import ssl
import os

from subprocess import check_output
from itertools import cycle
from functools import wraps
from time import time

__all__ = ['SockFeed', 'HTTPCons', 'unit_change', 'bar']


def bar(width=0, fill='#'):
    """
    进度条处理
    :param width: 手动设置进度条宽度
    :param fill: 进度填充字符
    """
    def function_wrapper(func):
        @wraps(func)
        def arguments(self, *args, **kwargs):
            if not hasattr(self, 'progressed') or not hasattr(self, 'total'):
                print("progressed, total attribute is needed!")
                return
            progress_cursor = 1
            last_update = 0
            while self.progressed <= self.total:
                func(self, *args, **kwargs)
                if not hasattr(self, 'disable_progress') or not self.disable_progress:
                    if self.total <= 0:
                        print("Total Length Invalid !")
                        self.progressed = self.total = 1
                        break
                    if time() - last_update > .1:
                        if not width:
                            try:
                                w = int(check_output("stty size", stderr=None, shell=True).split(b" ")[1])
                            except:
                                w = 50
                        else:
                            w = width
                        if not hasattr(self, 'chunked') or not self.chunked:
                            # 普通编码进度条
                            percent = self.progressed / float(self.total)
                            # marks count
                            percent_show = "{}%".format(int(percent * 100))
                            # marks width
                            title = getattr(self, 'title', '')
                            mark_width = w - len(percent_show) - str_len(title) - 7
                            mark_count = int(math.floor(mark_width * percent))
                            sys.stdout.write(
                                ' ' + title + ' ' +
                                '[' + fill * mark_count + ' ' * (mark_width - mark_count) + ']  ' + percent_show + '\r')
                        else:
                            # 分块编码进度条
                            progress_cursor += 1
                            title = getattr(self, 'title', '')
                            chunk_recved = unit_change(self.chunk_recved)
                            mark_width = w - str_len(title) - len(chunk_recved) - 6
                            sys.stdout.write(" " + title + " " +
                                             "[" +
                                             "".join([i for _, i in zip(range(mark_width),
                                                                        cycle([">> ", " >>", "> >"][progress_cursor % 3]))])
                                             + "] {}\r".format(chunk_recved))

                        sys.stdout.flush()
                        last_update = time()

                    if self.progressed == self.total:
                        sys.stdout.write(" " * w + '\r')
                        break
                else:
                    if self.progressed == self.total:
                        break
        return arguments
    return function_wrapper


def unit_change(target):
    """
    单位换算
    :param target: unsigned int
    :return: str
    """
    if target < 0:
        return str(target)
    unit_list = ('B', 'KB', 'MB', 'GB', 'TB')
    index = 0
    target = float(target)
    while target > 1024:
        index += 1
        target /= 1024
    return "{:.2f} {}".format(round(target, 2), unit_list[index])


def str_len(s):
    """
    获取占用等宽字体终端实际宽度，适用`Monaco`等其他等宽字体字体
    :param s:
    :return:
    """
    length = 0
    for i in s:
        # Chinese,Japanese,Korean character utf8 range
        # Test Font `Monaco`
        if 3105 <= ord(i) <= 65535:
            length += 2
        else:
            length += 1
    return length


class SockFeed(object):
    """
    连接响应
    """
    def __init__(self, connection):
        """
        :param connection: HttpCons
        """
        self.con = connection
        self.socket = self.con.connect
        self.status = None
        self.raw_head = b''
        self.headers = {}
        self.data = b''
        self.progressed = 0
        self.total = 0
        self.disable_progress = False
        self.chunked = False
        self.current_chunk = b''
        self.chunk_recved = 0
        self.title = ''

        self.file_handle = None

    def finish_loop(self):
        """
        让进度条走满，关闭TCP连接，关闭可能还打开的文件
        :return: None
        """
        self.progressed = self.total = 100
        self.con.close()  # 关闭tcp连接
        if self.file_handle:
            self.file_handle.close()

    def clean_failed_file(self):
        """
        下载失败后清理、删除文件
        :return:
        """
        if self.file_handle:
            name = self.file_handle.name
            self.file_handle.close()
            os.unlink(name)

    def save_data(self, data):
        """
        将每次获取的HTTP实体保存进内存或文件
        :param data:
        :return:
        """
        if self.file_handle:
            self.file_handle.write(data)
        else:
            self.data += data

    def flush_chunk(self, data):
        self.current_chunk += data

        while len(self.current_chunk) > 10240 or self.current_chunk.endswith(b'0\r\n\r\n'):  # 并不意味着所有分块结束
            # 开始解析当前chunk cache
            chunk_head = self.current_chunk[: self.current_chunk.index(b'\r\n')]
            chunk_left = self.current_chunk[self.current_chunk.index(b'\r\n') + 2:]
            chunk_size = int(chunk_head, 16)
            if chunk_size == 0:
                self.finish_loop()  # 一定要用finish_loop结束请求，否则会出现未关闭的文件 !
                return True
            if chunk_size > len(chunk_left):
                # 说明当前分块没有接收完全
                return False
            valid_data = chunk_left[: chunk_size]
            self.chunk_recved += len(valid_data)
            self.save_data(valid_data)
            self.current_chunk = chunk_left[chunk_size:]
            if self.current_chunk.startswith(b'\r\n'):
                # 如果上一个分块没有吃掉最后的 \r\n，则在这里把它剔除
                self.current_chunk = self.current_chunk[2:]

    @bar()
    def http_response(self, file_path='', skip_body=False, chunk=4096, overwrite=False):
        """
        通过进度条控制获取响应结果
        :param file_path: str => 下载文件位置，若文件已存在，则在前面用数字区分版本
        :param skip_body: bool => 是否跳过http实体
        :param chunk: int => 缓存块大小
        :param overwrite: bool => 是否覆盖重名文件
        :return:
        """
        if file_path and not self.file_handle:
            path_choice = file_path
            if os.path.exists(file_path):
                if overwrite:
                    os.remove(file_path)
                else:
                    file_index = 1
                    dirname = os.path.dirname(path_choice)
                    filename = os.path.basename(path_choice)
                    while os.path.exists(path_choice):
                        path_choice = os.path.join(dirname, '{}_{}'.format(file_index, filename))
                        file_index += 1

            self.file_handle = open(path_choice, 'wb')
            self.title = os.path.basename(path_choice)

        if self.status and self.progressed == self.total:
            self.finish_loop()
            return True

        data = self.socket.recv(chunk)

        if not data:
            self.finish_loop()
            return True
        if not self.status:
            self.raw_head += data
            if b'\r\n\r\n' in self.raw_head:  # 接收数据直到 `\r\n\r\n` 为止
                seps = self.raw_head[0: self.raw_head.index(b'\r\n\r\n')].split(b'\r\n')
                status = seps[0].split(b' ')
                self.status = {
                    'status': seps[0],
                    'code': status[1],
                    'version': status[0]
                }
                if self.file_handle and not self.status['code'] == b'200':
                    self.clean_failed_file()
                    self.finish_loop()
                    return False
                self.headers = {
                    i.split(b":")[0]: i.split(b":")[1].strip() for i in seps[1:]
                }
                # print("\n".join(["{} => {}".format(str(k), str(self.headers[k])) for k in self.headers]))
                if skip_body:
                    self.finish_loop()
                    return True

                if b'Content-Length' in self.headers:
                    self.total = int(self.headers[b'Content-Length'])
                else:
                    self.total = 100
                    self.chunked = True

                left = self.raw_head[self.raw_head.index(b"\r\n\r\n") + 4:]

                if left:
                    if not self.chunked:
                        self.save_data(left)
                        self.progressed += len(left)
                    else:

                        self.flush_chunk(left)  # 实体部分以分块大小十六进制数字开头

        else:
            if not self.chunked:
                self.save_data(data)
                self.progressed += len(data)
                if self.progressed == self.total:
                    self.finish_loop()
            else:
                self.flush_chunk(data)


class HTTPCons(object):
    """
    启动连接，发出请求
    """
    def __init__(self, debug=False):
        self.host = ''
        self.port = 0
        self.is_debug = debug
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connect = None

    def __del__(self):
        self.close()

    def close(self):
        """
        需要请求接收完成之后手动关闭
        :return: None
        """
        if self.connect is self.s:
            self.connect.close()
        else:
            self.connect.close()
            self.s.close()

    def https_init(self, host, port):
        """
        https连接
        :param host: str
        :param port: int
        :return: None
        """
        context = ssl.create_default_context()
        context.check_hostname = True
        context.verify_mode = ssl.CERT_REQUIRED
        context.load_default_certs()
        self.connect = context.wrap_socket(self.s, server_hostname=host)
        # self.connect.settimeout(65)
        self.connect.connect((host, port))
        self.host = host
        self.port = port

    def http_init(self, host, port):
        """
        http连接
        :param host: str
        :param port: int
        :return: None
        """
        self.connect = self.s
        # self.connect.settimeout(60)
        self.connect.connect((host, port))
        self.host = host
        self.port = port

    def request(self, url, method='GET', headers=None, data=None):
        """
        链接解析，完成请求
        :param url: str
        :param method: str => GET | POST
        :param headers: dict
        :param data: str => post data entity
        :return: None
        """
        if '//' not in url:
            raise URLNotComplete(url, 'url protocol')
        index = url.index('//')
        ishttps = url[0: index - 1].lower() == 'https'
        host_url = url[index + 2:]
        if "/" not in host_url:
            host_url += '/'
        index = host_url.index("/")
        host_port = host_url[0: index]
        url = host_url[index:]
        port = None
        if ':' in host_port:
            split = host_port.split(":")
            host = split[0]
            port = int(split[1].split("/")[0])
        else:
            host = host_port

        if ishttps:
            if not port:
                port = 443
            self.https_init(host, port)
        else:
            if not port:
                port = 80
            self.http_init(host, port)
        self.__send(url, method, headers, post_data=data)
        return self.connect

    def __send(self, href, method='GET', headers=None, post_data=None):
        data = """{method} {href} HTTP/1.1\r\n{headers}\r\n\r\n"""
        UA = "TerminalPrinter http client"
        if not headers:
            head = """Host: {}\r\n""".format(self.host)
            head += "User-Agent: " + UA
        else:
            head = "\r\n".join(["{}: {}".format(x, headers[x]) for x in headers])
            if 'Host' not in headers:
                head += """\r\nHost: {}""".format(self.host)
            if 'User-Agent' not in headers:
                head += "\r\nUser-Agent: " + UA
        if method == 'POST':
            if data and type(data) == str:
                # upload for one time
                head += "\r\nContent-Length: {}".format(len(post_data))
                head += "\r\n\r\n{}".format(post_data)
            else:
                raise URLNotComplete(href, 'POST data')
        elif method == 'GET':
            if post_data:
                if not type(post_data) == dict:
                    raise Exception("post data must be a dict")
                if '?' not in href[-1]:
                    href += '?'

                for i in post_data:
                    href += '{}={}&'.format(i, post_data[i])
        data = data.format(method=method, href=href, headers=head)
        if method == 'POST':
            data = data[:-4]
        if self.is_debug:
            print("\033[01;33mRequest:\033[00m\033[01;31m(DANGER)\033[00m")
            print(data.__repr__().strip("'"))
        if sys.version_info.major == 3:
            self.connect.sendall(data.encode())
        else:
            self.connect.sendall(data)


class URLNotComplete(Exception):
    def __init__(self, url, lack):
        self.url = url
        self.lack = lack

    def __str__(self):
        return "URL: {} missing {}".format(self.url, self.lack)


if __name__ == '__main__':
    import unittest
    import hashlib
    import tempfile

    class HTTPTest(unittest.TestCase):
        """
        static.hellflame.net 域名下的文件大多数情况下都是chunked编码
        raw.githubusercontent.com 域名下文件未分块
        """
        def test_https_request(self):
            req = HTTPCons()
            connect = req.request("https://static.hellflame.net/resource/de5ca9cf5320673dc43b526e3d737f05")
            self.assertEqual(req.host, 'static.hellflame.net')
            self.assertEqual(req.port, 443)
            self.assertIs(connect, req.connect)
            req.close()

        def test_http_request(self):
            req = HTTPCons()
            connect = req.request("http://static.hellflame.net/resource/de5ca9cf5320673dc43b526e3d737f05")
            self.assertEqual(req.host, 'static.hellflame.net')
            self.assertEqual(req.port, 80)
            self.assertIs(connect, req.connect)
            req.close()

        def test_response_in_memory(self):
            req = HTTPCons()
            req.request("https://static.hellflame.net/resource/c8c12b1c34af9808c34fa60d862016b7")
            resp = SockFeed(req)
            resp.disable_progress = True
            resp.http_response()
            self.assertEqual(hashlib.md5(resp.data).hexdigest(), '9a50ddbef4c82eb9003bd496a00e0989')

        def test_response_downloading(self):
            file_path = os.path.join(tempfile.gettempdir(), '1m.data')
            req = HTTPCons()
            req.request("https://static.hellflame.net/resource/c8c12b1c34af9808c34fa60d862016b7")
            resp = SockFeed(req)
            resp.disable_progress = True
            resp.http_response(file_path, overwrite=True)

            with open(file_path, 'rb') as handle:
                content = handle.read()

            os.remove(resp.file_handle.name)
            self.assertEqual(hashlib.md5(content).hexdigest(), '9a50ddbef4c82eb9003bd496a00e0989')

        def test_small_response_in_memory(self):
            req = HTTPCons()
            req.request("https://static.hellflame.net/resource/5573012afe7227ab4457331df42af57d")
            resp = SockFeed(req)
            resp.disable_progress = True
            resp.http_response()
            self.assertEqual(hashlib.md5(resp.data).hexdigest(), '8688229badcaa3cb2730dab99a618be6')

        def test_small_response_downloading(self):
            file_path = os.path.join(tempfile.gettempdir(), '3k.data')
            req = HTTPCons()
            req.request("https://static.hellflame.net/resource/5573012afe7227ab4457331df42af57d")
            resp = SockFeed(req)
            resp.disable_progress = True
            resp.http_response(file_path, overwrite=True)
            with open(file_path, 'rb') as handle:
                content = handle.read()
            os.remove(resp.file_handle.name)
            self.assertEqual(hashlib.md5(content).hexdigest(), '8688229badcaa3cb2730dab99a618be6')

        def test_non_chunked_in_memory(self):
            req = HTTPCons()
            req.request("https://raw.githubusercontent.com/hellflame/qiniu_manager/v1.4.6/qiniuManager/manager.py")
            resp = SockFeed(req)
            resp.disable_progress = True
            resp.http_response()
            self.assertEqual(hashlib.md5(resp.data).hexdigest(), '276efce035d49f7f3ea168b720075523')


    unittest.main(verbosity=2)
